#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from glsl_math import *
from polygon_sampling_ported import *
from cubic_solver_ported import solve_cubic

MIN_POLYGONAL_LIGHT_VERTEX_COUNT = 3
MAX_POLYGONAL_LIGHT_VERTEX_COUNT = 4
MAX_POLYGON_VERTEX_COUNT = 5




def sample_area_polygon_turk(vertex_count, vertices, fan_areas, random_numbers):
    target_area = fan_areas[MAX_POLYGON_VERTEX_COUNT - 3][1] * random_numbers[0]
    subtriangle_area = target_area
    triangle_area = fan_areas[0][0]
    triangle_vertices = [
        vertices[1], vertices[0], vertices[2]
    ]
    for i in range(0, MAX_POLYGON_VERTEX_COUNT - 3):
        if (i + 3 >= vertex_count or fan_areas[i][1] >= target_area):
            break
        subtriangle_area = target_area - fan_areas[i][1]
        triangle_area = fan_areas[i + 1][0]
        triangle_vertices[0] = vertices[i + 2]
        triangle_vertices[2] = vertices[i + 3]
    random_numbers[0] = subtriangle_area / triangle_area
    sqrt_random_0 = sqrt(random_numbers[0])
    barycentric = vec3(
        1.0 - sqrt_random_0,
        sqrt_random_0 * random_numbers[1],
        fma(-sqrt_random_0, random_numbers[1], sqrt_random_0)
    )
    return barycentric[0] * triangle_vertices[0] + barycentric[1] * triangle_vertices[1] + barycentric[2] * triangle_vertices[2]


def get_area_sample_density(out_normalized_dir, light_sample, shading_position, light_normal, light_area):
    out_normalized_dir = light_sample - shading_position
    distance_squared = dot(out_normalized_dir, out_normalized_dir)
    normalization = inversesqrt(distance_squared)
    out_normalized_dir *= normalization
    projected_area = abs(dot(light_normal, out_normalized_dir)) * light_area
    return distance_squared / projected_area


class solid_angle_rectangle_urena_t:
    def __init__(self):
        self.solid_angle = 0.0


def prepare_solid_angle_rectangle_sampling_urena(s, ex, ey, exl, eyl, local_to_world_space, o):
    squad = solid_angle_rectangle_urena_t()
    squad.o = o
    squad[0] = local_to_world_space[0]
    squad[1] = local_to_world_space[1]
    squad[2] = local_to_world_space[2]
    d = s - o
    squad.z0 = dot(d, squad[2])
    squad[2] = where((squad.z0 > 0) ,-squad[2] ,squad[2])
    squad.z0 = -abs(squad.z0)
    squad.z0sq = squad.z0 * squad.z0
    squad.x0 = dot(d, squad[0])
    squad.y0 = dot(d, squad[1])
    squad.x1 = squad.x0 + exl
    squad.y1 = squad.y0 + eyl
    squad.y0sq = squad.y0 * squad.y0
    squad.y1sq = squad.y1 * squad.y1
    v00 = vec3(squad.x0, squad.y0, squad.z0)
    v01 = vec3(squad.x0, squad.y1, squad.z0)
    v10 = vec3(squad.x1, squad.y0, squad.z0)
    v11 = vec3(squad.x1, squad.y1, squad.z0)
    n0 = normalize(cross(v00, v10))
    n1 = normalize(cross(v10, v11))
    n2 = normalize(cross(v11, v01))
    n3 = normalize(cross(v01, v00))
    g0 = arccos(-dot(n0,n1))
    g1 = arccos(-dot(n1,n2))
    g2 = arccos(-dot(n2,n3))
    g3 = arccos(-dot(n3,n0))
    squad.b0 = n0[2]
    squad.b1 = n2[2]
    squad.b0sq = squad.b0 * squad.b0
    squad.k = 2.0 * pi - g2 - g3
    squad.solid_angle = g0 + g1 - squad.k
    return squad


def sample_solid_angle_rectangle_urena(squad, random_numbers):
    u = random_numbers[0]
    v = random_numbers[1]
    au = fma(u, squad.solid_angle, squad.k)
    fu = fma(cos(au), squad.b0, -squad.b1) / sin(au)
    cu = inversesqrt(fma(fu, fu, squad.b0sq))
    cu = where((fu > 0.0) ,cu ,-cu)
    cu = clip(cu, -1.0, 1.0)
    xu = -(cu * squad.z0) * inversesqrt(fma(-cu, cu, 1.0))
    xu = clip(xu, squad.x0, squad.x1)
    d = sqrt(xu*xu + squad.z0sq)
    h0 = squad.y0 * inversesqrt(fma(d, d, squad.y0sq))
    h1 = squad.y1 * inversesqrt(fma(d, d, squad.y1sq))
    hv = h0 + v * (h1 - h0)
    mhv2_1 = fma(-hv, hv, 1.0)
    yv = where((mhv2_1 >= 0.0) ,((hv * d) * inversesqrt(mhv2_1)) ,squad.y1)
    return normalize(xu * squad[0] + yv * squad[1] + squad.z0 * squad[2])


class solid_angle_polygon_arvo_t:
    def __init__(self):
        self.vertex_count = 0
        self.vertex_dirs = [zeros((3,)) for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.fan_solid_angles = [0.0 for _i in range(MAX_POLYGON_VERTEX_COUNT - 2)]
        self.opposite_dirs = [zeros((2,)) for _i in range(MAX_POLYGON_VERTEX_COUNT - 2)]
        self.solid_angle = 0.0


def prepare_solid_angle_polygon_sampling_arvo(vertex_count, vertices, shading_position):
    polygon = solid_angle_polygon_arvo_t()
    for i in range(0, MAX_POLYGON_VERTEX_COUNT):
        polygon.vertex_dirs[i] = normalize(vertices[i] - shading_position)
    solid_angle = 0.0
    for i in range(0, MAX_POLYGON_VERTEX_COUNT - 2):
        if (i >= 1 and i + 2 >= vertex_count):
            break
        edge_normals = [
            normalize(cross(polygon.vertex_dirs[i + 1] - polygon.vertex_dirs[0], polygon.vertex_dirs[0])),
            normalize(cross(polygon.vertex_dirs[i + 2] - polygon.vertex_dirs[i + 1], polygon.vertex_dirs[i + 1])),
        ]
        polygon.opposite_dirs[i][0] = -dot(edge_normals[0], edge_normals[1])
        polygon.opposite_dirs[i][1] = sqrt(maximum(0.0, fma(-polygon.opposite_dirs[i][0], polygon.opposite_dirs[i][0], 1.0)))
        dot_0_1 = dot(polygon.vertex_dirs[0], polygon.vertex_dirs[i + 1])
        dot_0_2 = dot(polygon.vertex_dirs[0], polygon.vertex_dirs[i + 2])
        dot_1_2 = dot(polygon.vertex_dirs[i + 1], polygon.vertex_dirs[i + 2])
        simplex_volume = linalg.det(mat3(polygon.vertex_dirs[0], polygon.vertex_dirs[i + 1], polygon.vertex_dirs[i + 2]))
        tangent = abs(simplex_volume) / (1.0 + dot_0_1 + dot_0_2 + dot_1_2)
        solid_angle += 2.0 * positive_atan(tangent)
        polygon.fan_solid_angles[i] = solid_angle
    polygon.solid_angle = solid_angle
    polygon.vertex_count = vertex_count
    return polygon


def sample_solid_angle_polygon_arvo(polygon, random_numbers):
    target_solid_angle = polygon.solid_angle * random_numbers[0]
    subtriangle_solid_angle = target_solid_angle
    opposite_dir = polygon.opposite_dirs[0]
    triangle_vertices = [
        polygon.vertex_dirs[1], polygon.vertex_dirs[0], polygon.vertex_dirs[2]
    ]
    for i in range(0, MAX_POLYGON_VERTEX_COUNT - 3):
        if (i + 3 >= polygon.vertex_count or polygon.fan_solid_angles[i] >= target_solid_angle):
            break
        subtriangle_solid_angle = target_solid_angle - polygon.fan_solid_angles[i]
        triangle_vertices[0] = polygon.vertex_dirs[i + 2]
        triangle_vertices[2] = polygon.vertex_dirs[i + 3]
        opposite_dir = polygon.opposite_dirs[i + 1]
    subtriangle_solid_angle_dir = vec2(cos(subtriangle_solid_angle), sin(subtriangle_solid_angle))
    p = subtriangle_solid_angle_dir[1] * opposite_dir[0] - subtriangle_solid_angle_dir[0] * opposite_dir[1]
    q = subtriangle_solid_angle_dir[1] * opposite_dir[1] + subtriangle_solid_angle_dir[0] * opposite_dir[0]
    u = q - opposite_dir[0]
    v = p + opposite_dir[1] * dot(triangle_vertices[0], triangle_vertices[1])
    s = ((v * q - u * p) * opposite_dir[0] - v) / ((v * p + u * q) * opposite_dir[1])
    edge_tangent_2_0 = normalize(triangle_vertices[2] - dot(triangle_vertices[0], triangle_vertices[2]) * triangle_vertices[0])
    vertex_2 = s * triangle_vertices[0] + sqrt(clip(fma(-s, s, 1.0), 0.0, 1.0)) * edge_tangent_2_0
    z = 1.0 - random_numbers[1] * (1.0 - dot(vertex_2, triangle_vertices[1]))
    edge_tangent_2_1 = normalize(vertex_2 - dot(triangle_vertices[1], vertex_2) * triangle_vertices[1])
    return z * triangle_vertices[1] + sqrt(clip(fma(-z, z, 1.0), 0.0, 1.0)) * edge_tangent_2_1


class bilinear_cosine_warp_polygon_hart_t:
    def __init__(self):
        self.polygon = solid_angle_polygon_t()
        self.density_0 = 0.0
        self.density_1 = zeros((2,))


def prepare_bilinear_cosine_warp_polygon_sampling_hart(vertex_count, vertices):
    polygon = bilinear_cosine_warp_polygon_hart_t()
    polygon.polygon = prepare_solid_angle_polygon_sampling(vertex_count, vertices, vec3(0.0))
    polygon.density_0 = maximum(0.0, polygon.polygon.vertex_dirs[0][2])
    polygon.density_1[0] = maximum(0.0, polygon.polygon.vertex_dirs[1][2])
    polygon.density_1[1] = polygon.polygon.vertex_dirs[2][2]
    for i in range(3, MAX_POLYGON_VERTEX_COUNT):
        polygon.density_1[1] = where((i < vertex_count) ,polygon.polygon.vertex_dirs[i][2] ,polygon.density_1[1])
    polygon.density_1[1] = maximum(0.0, polygon.density_1[1])
    density_sum = 2.0 * polygon.density_0 + polygon.density_1[0] + polygon.density_1[1]
    normalization = 4.0 / (polygon.polygon.solid_angle * density_sum)
    polygon.density_0 *= normalization
    polygon.density_1 *= normalization
    inv_solid_angle = 1.0 / polygon.polygon.solid_angle
    polygon.density_0 = where((density_sum <= 0.0) ,inv_solid_angle ,polygon.density_0)
    polygon.density_1 = where((density_sum <= 0.0) ,vec2(inv_solid_angle) ,polygon.density_1)
    return polygon


def linear_warp(random_number, density_0, density_1):
    lerped_density_sq = mix_fma(density_0 * density_0, density_1 * density_1, random_number)
    divisor = density_0 + sqrt(lerped_density_sq)
    return random_number * (density_0 + density_1) / divisor


def sample_bilinear_cosine_warp_polygon_hart(out_density, polygon, random_numbers):
    random_numbers[1] = linear_warp(random_numbers[1], 2.0 * polygon.density_0, dot(polygon.density_1, vec2(1.0)))
    density_0 = mix_fma(polygon.density_0, polygon.density_1[0], random_numbers[1])
    density_1 = mix_fma(polygon.density_0, polygon.density_1[1], random_numbers[1])
    random_numbers[0] = linear_warp(random_numbers[0], density_0, density_1)
    out_density = mix_fma(density_0, density_1, random_numbers[0])
    return sample_solid_angle_polygon(polygon.polygon, random_numbers)


class biquadratic_cosine_warp_polygon_hart_t:
    def __init__(self):
        self.polygon = solid_angle_polygon_t()
        self.density_0 = 0.0
        self.density_1 = zeros((3,))
        self.density_2 = zeros((3,))


def prepare_biquadratic_cosine_warp_polygon_sampling_hart(vertex_count, vertices):
    polygon = biquadratic_cosine_warp_polygon_hart_t()
    polygon.polygon = prepare_solid_angle_polygon_sampling(vertex_count, vertices, vec3(0.0))
    last_vertex = polygon.polygon.vertex_dirs[2]
    for i in range(3, MAX_POLYGON_VERTEX_COUNT):
        last_vertex = where((i < vertex_count) ,polygon.polygon.vertex_dirs[i] ,last_vertex)
    vertex_0 = polygon.polygon.vertex_dirs[0]
    polygon.density_0 = maximum(0.0, vertex_0[2])
    polygon.density_2[0] = maximum(0.0, polygon.polygon.vertex_dirs[1][2])
    polygon.density_2[2] = maximum(0.0, last_vertex[2])
    sample_2_1 = sample_solid_angle_polygon(polygon.polygon, vec2(0.5, 1.0))
    polygon.density_2[1] = maximum(0.0, sample_2_1[2])
    far_vertices = [ vertex_0, sample_2_1, last_vertex ]
    for i in range(0, 3):
        s2 = dot(vertex_0, far_vertices[i])
        s = fma(0.5, s2, 0.5)
        t = sqrt(maximum(0.0, fma(-s, s, 1.0)))
        t_axis_z = fma(-s2, vertex_0[2], far_vertices[i][2])
        normalization_t_axis = inversesqrt(2.0 * fma(-s2, s2, 1.0))
        sample_1_i_z = s * vertex_0[2] + (t * normalization_t_axis) * t_axis_z
        polygon.density_1[i] = maximum(0.0, sample_1_i_z)
    density_sum = 3.0 * polygon.density_0 + dot(polygon.density_1, vec3(1.0)) + dot(polygon.density_2, vec3(1.0))
    normalization = 9.0 / (polygon.polygon.solid_angle * density_sum)
    polygon.density_0 *= normalization
    polygon.density_1 *= normalization
    polygon.density_2 *= normalization
    inv_solid_angle = 1.0 / polygon.polygon.solid_angle
    polygon.density_0 = where((density_sum <= 0.0) ,inv_solid_angle ,polygon.density_0)
    polygon.density_1 = where((density_sum <= 0.0) ,vec3(inv_solid_angle) ,polygon.density_1)
    polygon.density_2 = where((density_sum <= 0.0) ,vec3(inv_solid_angle) ,polygon.density_2)
    return polygon


def quadratic_warp(random_number, density_0, density_1, density_2):
    quadratic = vec3(density_0, 2.0 * (density_1 - density_0), density_0 - 2.0 * density_1 + density_2)
    cubic = vec4(0.0, quadratic[0], 0.5 * quadratic[1], (1.0 / 3.0) * quadratic[2])
    random_number *= dot(cubic.yzw, vec3(1.0))
    cubic[0] = -random_number
    roots = zeros((3,))
    if (solve_cubic(roots, cubic)):
        result = roots[0]
        result = where((roots[1] >= 0.0 and roots[1] <= 1.0) ,roots[1] ,result)
        result = where((roots[2] >= 0.0 and roots[2] <= 1.0) ,roots[2] ,result)
        return result
    else:
        return roots[0]


def quadratic_bezier(b_0_0, b_0_1, b_0_2, location):
    b_1_0 = mix_fma(b_0_0, b_0_1, location)
    b_1_1 = mix_fma(b_0_1, b_0_2, location)
    return mix_fma(b_1_0, b_1_1, location)


def sample_biquadratic_cosine_warp_polygon_hart(out_density, polygon, random_numbers):
    random_numbers[1] = quadratic_warp(random_numbers[1], 3.0 * polygon.density_0, dot(polygon.density_1, vec3(1.0)), dot(polygon.density_2, vec3(1.0)))
    density_0 = quadratic_bezier(polygon.density_0, polygon.density_1[0], polygon.density_2[0], random_numbers[1])
    density_1 = quadratic_bezier(polygon.density_0, polygon.density_1[1], polygon.density_2[1], random_numbers[1])
    density_2 = quadratic_bezier(polygon.density_0, polygon.density_1[2], polygon.density_2[2], random_numbers[1])
    random_numbers[0] = quadratic_warp(random_numbers[0], density_0, density_1, density_2)
    out_density = quadratic_bezier(density_0, density_1, density_2, random_numbers[0])
    return sample_solid_angle_polygon(polygon.polygon, random_numbers)


class edge_arvo_t:
    def __init__(self):
        self.cdf_factor = 0.0
        self.length_coeffs = zeros((2,))
        self.elevations = zeros((2,))


class projected_solid_angle_polygon_arvo_t:
    def __init__(self):
        self.vertex_count = 0
        self.vertex_azimuths = [0.0 for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.edges = [edge_arvo_t() for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.inner_edge_0 = edge_arvo_t()
        self.sector_projected_solid_angles = [0.0 for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.projected_solid_angle = 0.0


def prepare_edge_arvo(vertex_0, vertex_1):
    edge = edge_arvo_t()
    normal_a = normalize(cross(vertex_0, vertex_1))
    edge.cdf_factor = 0.5 * normal_a[2]
    ccw_vertex = where((edge.cdf_factor > 0.0) ,vertex_0 ,vertex_1)
    normal_c = rotate_90(normalize(ccw_vertex[0:2]))
    cos_beta = -dot(normal_a[0:2], normal_c)
    sin_beta_sq = fma(-cos_beta, cos_beta, 1.0)
    csc_beta = inversesqrt(maximum(0.0, sin_beta_sq))
    csc_c = inversesqrt(maximum(0.0, fma(-ccw_vertex[2], ccw_vertex[2], 1.0)))
    edge.length_coeffs[0] = sin_beta_sq
    edge.length_coeffs[1] = dot(normal_a[0:2], rotate_90(normal_c)) * cos_beta
    edge.length_coeffs *= csc_beta * csc_c
    edge.elevations[0] = ccw_vertex[2]
    edge.elevations[1] = cross(ccw_vertex, normal_a)[2]
    edge.elevations[1] = where((edge.cdf_factor > 0.0) ,-edge.elevations[1] ,edge.elevations[1])
    return edge


def get_edge_projected_solid_angle_in_sector_arvo(edge, relative_azimuth_0, relative_azimuth_1):
    dir_0 = vec2(cos(relative_azimuth_0), sin(relative_azimuth_0))
    point_0 = vec2(dot(edge.length_coeffs, dir_0), dir_0[1])
    dir_1 = vec2(cos(relative_azimuth_1), sin(relative_azimuth_1))
    point_1 = vec2(dot(edge.length_coeffs, dir_1), dir_1[1])
    rotated_point = vec2(
        point_0[0] * point_1[0] + point_0[1] * point_1[1],
        point_0[0] * point_1[1] - point_0[1] * point_1[0])
    length = positive_atan(abs(rotated_point[1]) / rotated_point[0])
    return edge.cdf_factor * length


def get_edge_projected_solid_angle_in_sector_derivative_arvo(edge, relative_azimuth_0, relative_azimuth_1):
    dir_0 = vec2(cos(relative_azimuth_0), sin(relative_azimuth_0))
    point_0 = vec2(dot(edge.length_coeffs, dir_0), dir_0[1])
    dir_1 = vec2(cos(relative_azimuth_1), sin(relative_azimuth_1))
    point_1 = vec2(dot(edge.length_coeffs, dir_1), dir_1[1])
    rotated_point = vec2(
        point_0[0] * point_1[0] + point_0[1] * point_1[1],
        point_0[0] * point_1[1] - point_0[1] * point_1[0])
    quotient = abs(rotated_point[1]) / rotated_point[0]
    length = positive_atan(quotient)
    dir_1_deriv = rotate_90(dir_1)
    point_1_deriv = vec2(dot(edge.length_coeffs, dir_1_deriv), dir_1_deriv[1])
    rotated_point_deriv = vec2(
        point_0[0] * point_1_deriv[0] + point_0[1] * point_1_deriv[1],
        point_0[0] * point_1_deriv[1] - point_0[1] * point_1_deriv[0])
    quotient_derivative = (rotated_point_deriv[1] * rotated_point[0] - rotated_point[1] * rotated_point_deriv[0]) / (rotated_point[0] * rotated_point[0])
    quotient_derivative = where((rotated_point[1] < 0.0) ,(-quotient_derivative) ,quotient_derivative)
    length_deriv = quotient_derivative / fma(quotient, quotient, 1.0)
    return edge.cdf_factor * vec2(length, length_deriv)


def get_edge_elevation_arvo(edge, relative_azimuth):
    dir = vec2(cos(relative_azimuth), sin(relative_azimuth))
    point = vec2(dot(edge.length_coeffs, dir), dir[1])
    point = normalize(point)
    return dot(point, edge.elevations)


def compare_and_swap_arvo(polygon, lhs, rhs):
    lhs_azimuth = polygon.vertex_azimuths[lhs]
    flip = polygon.vertex_azimuths[lhs] - polygon.vertex_azimuths[rhs]
    polygon.vertex_azimuths[lhs] = where((flip > 0.0) ,polygon.vertex_azimuths[rhs] ,lhs_azimuth)
    polygon.vertex_azimuths[rhs] = where((flip > 0.0) ,lhs_azimuth ,polygon.vertex_azimuths[rhs])
    lhs_edge = polygon.edges[lhs]
    polygon.edges[lhs] = where((flip > 0.0) ,polygon.edges[rhs] ,lhs_edge)
    polygon.edges[rhs] = where((flip > 0.0) ,lhs_edge ,polygon.edges[rhs])


def sort_convex_polygon_vertices_arvo(polygon):
    if (polygon.vertex_count == 3):
        compare_and_swap_arvo(polygon, 1, 2)
#if MAX_POLYGON_VERTEX_COUNT >= 4
    elif (polygon.vertex_count == 4):
        compare_and_swap_arvo(polygon, 1, 3)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 5
    elif (polygon.vertex_count == 5):
        compare_and_swap_arvo(polygon, 2, 4)
        compare_and_swap_arvo(polygon, 1, 3)
        compare_and_swap_arvo(polygon, 1, 2)
        compare_and_swap_arvo(polygon, 0, 3)
        compare_and_swap_arvo(polygon, 3, 4)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 6
    elif (polygon.vertex_count == 6):
        compare_and_swap_arvo(polygon, 3, 5)
        compare_and_swap_arvo(polygon, 2, 4)
        compare_and_swap_arvo(polygon, 1, 5)
        compare_and_swap_arvo(polygon, 0, 4)
        compare_and_swap_arvo(polygon, 4, 5)
        compare_and_swap_arvo(polygon, 1, 3)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 7
    elif (polygon.vertex_count == 7):
        compare_and_swap_arvo(polygon, 2, 5)
        compare_and_swap_arvo(polygon, 1, 6)
        compare_and_swap_arvo(polygon, 5, 6)
        compare_and_swap_arvo(polygon, 3, 4)
        compare_and_swap_arvo(polygon, 0, 4)
        compare_and_swap_arvo(polygon, 4, 6)
        compare_and_swap_arvo(polygon, 1, 3)
        compare_and_swap_arvo(polygon, 3, 5)
        compare_and_swap_arvo(polygon, 4, 5)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 8
    elif (polygon.vertex_count == 8):
        compare_and_swap_arvo(polygon, 2, 6)
        compare_and_swap_arvo(polygon, 3, 7)
        compare_and_swap_arvo(polygon, 1, 5)
        compare_and_swap_arvo(polygon, 0, 4)
        compare_and_swap_arvo(polygon, 4, 6)
        compare_and_swap_arvo(polygon, 5, 7)
        compare_and_swap_arvo(polygon, 6, 7)
        compare_and_swap_arvo(polygon, 4, 5)
        compare_and_swap_arvo(polygon, 1, 3)
#endif
    compare_and_swap_arvo(polygon, 0, 2)
#if MAX_POLYGON_VERTEX_COUNT >= 4
    if (polygon.vertex_count >= 4):
        compare_and_swap_arvo(polygon, 2, 3)
#endif
    compare_and_swap_arvo(polygon, 0, 1)


def prepare_projected_solid_angle_polygon_sampling_arvo(vertex_count, vertices):
    polygon = projected_solid_angle_polygon_arvo_t()
    for i in range(0, MAX_POLYGON_VERTEX_COUNT):
        vertices[i] = normalize(vertices[i])
    polygon.vertex_count = vertex_count
    polygon.inner_edge_0.cdf_factor = 1.0
    polygon.inner_edge_0.length_coeffs = polygon.inner_edge_0.elevations = vec2(0.0)
    polygon.vertex_azimuths[0] = arctan2(vertices[0][1], vertices[0][0])
    polygon.edges[0] = prepare_edge_arvo(vertices[0], vertices[1])
    previous_edge = polygon.edges[0]
    for i in range(1, MAX_POLYGON_VERTEX_COUNT):
        polygon.vertex_azimuths[i] = arctan2(vertices[i][1], vertices[i][0])
        polygon.vertex_azimuths[i] -= where((polygon.vertex_azimuths[i] > polygon.vertex_azimuths[0] + pi) ,(2.0 * pi),0.0)
        polygon.vertex_azimuths[i] += where((polygon.vertex_azimuths[i] < polygon.vertex_azimuths[0] - pi) ,(2.0 * pi),0.0)
        if (i > 2 and i == polygon.vertex_count):
            break
        edge = prepare_edge_arvo(vertices[i], vertices[(i + 1) % MAX_POLYGON_VERTEX_COUNT])
        polygon.edges[i] = where((edge.cdf_factor >= 0.0) ,edge ,previous_edge)
        polygon.inner_edge_0 = where((previous_edge.cdf_factor < 0.0 and edge.cdf_factor >= 0.0) ,previous_edge ,polygon.inner_edge_0)
        previous_edge = edge
    edge = polygon.edges[0]
    polygon.edges[0] = where((edge.cdf_factor >= 0.0) ,edge ,previous_edge)
    polygon.inner_edge_0 = where((previous_edge.cdf_factor < 0.0 and edge.cdf_factor >= 0.0) ,previous_edge ,polygon.inner_edge_0)
    polygon.projected_solid_angle = 0.0
    if (polygon.inner_edge_0.cdf_factor > 0.0):
        for i in range(0, MAX_POLYGON_VERTEX_COUNT):
            if (i > 2 and i == polygon.vertex_count):
                break
            polygon.sector_projected_solid_angles[i] = get_edge_projected_solid_angle_in_sector_arvo(polygon.edges[i], 0.0, polygon.vertex_azimuths[(i + 1) % MAX_POLYGON_VERTEX_COUNT] - polygon.vertex_azimuths[i])
            polygon.projected_solid_angle += polygon.sector_projected_solid_angles[i]
    else:
        sort_convex_polygon_vertices_arvo(polygon)
        inner_edge = polygon.inner_edge_0
        inner_azimuth = polygon.vertex_azimuths[0]
        outer_edge = edge_arvo_t()
        outer_azimuth = polygon.vertex_azimuths[0]
        for i in range(0, MAX_POLYGON_VERTEX_COUNT - 1):
            if (i > 1 and i + 1 == polygon.vertex_count):
                break
            vertex_edge = polygon.edges[i]
            vertex_azimuth = polygon.vertex_azimuths[i]
            if (i == 0):
                outer_edge = vertex_edge
            else:
                inner_edge = where((vertex_edge.cdf_factor >= 0.0) ,inner_edge ,vertex_edge)
                inner_azimuth = where((vertex_edge.cdf_factor >= 0.0) ,inner_azimuth ,vertex_azimuth)
                outer_edge = where((vertex_edge.cdf_factor >= 0.0) ,vertex_edge ,outer_edge)
                outer_azimuth = where((vertex_edge.cdf_factor >= 0.0) ,vertex_azimuth ,outer_azimuth)
            polygon.sector_projected_solid_angles[i] = get_edge_projected_solid_angle_in_sector_arvo(
                outer_edge, polygon.vertex_azimuths[i] - outer_azimuth, polygon.vertex_azimuths[i + 1] - outer_azimuth)
            polygon.sector_projected_solid_angles[i] += get_edge_projected_solid_angle_in_sector_arvo(
                inner_edge, polygon.vertex_azimuths[i] - inner_azimuth, polygon.vertex_azimuths[i + 1] - inner_azimuth)
            polygon.projected_solid_angle += polygon.sector_projected_solid_angles[i]
    return polygon


def evaluate_cubic_interpolation_polynomial(sample_x, x, y):
    y01 = (y[0] - y[1]) / (x[0] - x[1])
    y12 = (y[1] - y[2]) / (x[1] - x[2])
    y23 = (y[2] - y[3]) / (x[2] - x[3])
    y012 = (y01 - y12) / (x[0] - x[2])
    y123 = (y12 - y23) / (x[1] - x[3])
    y0123 = (y012 - y123) / (x[0] - x[3])
    return fma(sample_x - x[0], fma(sample_x - x[1], fma(sample_x - x[2], y0123, y012), y01), y[0])


def sample_sector_within_edge(random_numbers, target_projected_solid_angle, outer_edge, outer_azimuth, azimuth_0, azimuth_1, iteration_count):
    azimuths = [
        azimuth_0, mix_fma(azimuth_0, azimuth_1, 1.0 / 3.0),
        mix_fma(azimuth_0, azimuth_1, 2.0 / 3.0), azimuth_1
    ]
    projected_solid_angles = [0.0 for _i in range(4)]
    for i in range(0, 4):
        projected_solid_angles[i] = get_edge_projected_solid_angle_in_sector_arvo(
            outer_edge, azimuth_0 - outer_azimuth, azimuths[i] - outer_azimuth)
    sampled_azimuth = evaluate_cubic_interpolation_polynomial(target_projected_solid_angle, projected_solid_angles, azimuths)
    for i in range(0, iteration_count):
        outer_psa = get_edge_projected_solid_angle_in_sector_derivative_arvo(
            outer_edge, azimuth_0 - outer_azimuth, sampled_azimuth - outer_azimuth)
        error = outer_psa[0] - target_projected_solid_angle
        derivative = outer_psa[1]
        sampled_azimuth -= error / derivative
        sampled_azimuth = clip(sampled_azimuth, azimuth_0, azimuth_1)
    sampled_dir = zeros((3,))
    sampled_dir[0] = cos(sampled_azimuth)
    sampled_dir[1] = sin(sampled_azimuth)
    outer_z = get_edge_elevation_arvo(outer_edge, sampled_azimuth - outer_azimuth)
    sampled_dir[2] = sqrt(mix_fma(1.0, outer_z * outer_z, random_numbers[1]))
    sampled_dir[0:2] *= sqrt(fma(-sampled_dir[2], sampled_dir[2], 1.0))
    return sampled_dir


def sample_sector_between_edges(random_numbers, target_projected_solid_angle, inner_edge, inner_azimuth, outer_edge, outer_azimuth, azimuth_0, azimuth_1, iteration_count):
    azimuths = [
        azimuth_0, mix_fma(azimuth_0, azimuth_1, 1.0 / 3.0),
        mix_fma(azimuth_0, azimuth_1, 2.0 / 3.0), azimuth_1
    ]
    projected_solid_angles = [0.0 for _i in range(4)]
    for i in range(0, 4):
        projected_solid_angles[i] = get_edge_projected_solid_angle_in_sector_arvo(
            outer_edge, azimuth_0 - outer_azimuth, azimuths[i] - outer_azimuth)
        projected_solid_angles[i] += get_edge_projected_solid_angle_in_sector_arvo(
            inner_edge, azimuth_0 - inner_azimuth, azimuths[i] - inner_azimuth)
    sampled_azimuth = evaluate_cubic_interpolation_polynomial(target_projected_solid_angle, projected_solid_angles, azimuths)
    for i in range(0, iteration_count):
        outer_psa = get_edge_projected_solid_angle_in_sector_derivative_arvo(
            outer_edge, azimuth_0 - outer_azimuth, sampled_azimuth - outer_azimuth)
        inner_psa = get_edge_projected_solid_angle_in_sector_derivative_arvo(
            inner_edge, azimuth_0 - inner_azimuth, sampled_azimuth - inner_azimuth)
        error = inner_psa[0] + outer_psa[0] - target_projected_solid_angle
        derivative = inner_psa[1] + outer_psa[1]
        sampled_azimuth -= error / derivative
        sampled_azimuth = clip(sampled_azimuth, azimuth_0, azimuth_1)
    sampled_dir = zeros((3,))
    sampled_dir[0] = cos(sampled_azimuth)
    sampled_dir[1] = sin(sampled_azimuth)
    inner_z = get_edge_elevation_arvo(inner_edge, sampled_azimuth - inner_azimuth)
    outer_z = get_edge_elevation_arvo(outer_edge, sampled_azimuth - outer_azimuth)
    sampled_dir[2] = sqrt(mix_fma(inner_z * inner_z, outer_z * outer_z, random_numbers[1]))
    sampled_dir[0:2] *= sqrt(fma(-sampled_dir[2], sampled_dir[2], 1.0))
    return sampled_dir


def sample_projected_solid_angle_polygon_arvo(polygon, random_numbers, iteration_count):
    target_projected_solid_angle = random_numbers[0] * polygon.projected_solid_angle
    outer_edge = edge_arvo_t()
    if (polygon.inner_edge_0.cdf_factor > 0.0):
        for i in range(0, MAX_POLYGON_VERTEX_COUNT):
            if ((i > 2 and i == polygon.vertex_count) or (i > 0 and target_projected_solid_angle < 0.0)):
                break
            sector_projected_solid_angle = polygon.sector_projected_solid_angles[i]
            target_projected_solid_angle -= sector_projected_solid_angle
            outer_edge = polygon.edges[i]
            outer_azimuth = polygon.vertex_azimuths[i]
            azimuth_1 = polygon.vertex_azimuths[(i + 1) % MAX_POLYGON_VERTEX_COUNT]
        azimuth_1 = where((azimuth_1 < outer_azimuth) ,(azimuth_1 + 2.0 * pi) ,azimuth_1)
        target_projected_solid_angle += sector_projected_solid_angle
        random_numbers[0] = target_projected_solid_angle / sector_projected_solid_angle
        random_numbers[0] = clip(random_numbers[0], 0.0, 1.0)
        return sample_sector_within_edge(random_numbers, target_projected_solid_angle, outer_edge, outer_azimuth, outer_azimuth, azimuth_1, iteration_count)
    else:
        inner_edge = polygon.inner_edge_0
        inner_azimuth = polygon.vertex_azimuths[0]
        for i in range(0, MAX_POLYGON_VERTEX_COUNT - 1):
            if ((i > 1 and i + 1 == polygon.vertex_count) or (i > 0 and target_projected_solid_angle < 0.0)):
                break
            sector_projected_solid_angle = polygon.sector_projected_solid_angles[i]
            target_projected_solid_angle -= sector_projected_solid_angle
            vertex_edge = polygon.edges[i]
            vertex_azimuth = polygon.vertex_azimuths[i]
            if (i == 0):
                outer_edge = vertex_edge
                outer_azimuth = vertex_azimuth
            else:
                inner_edge = where((vertex_edge.cdf_factor >= 0.0) ,inner_edge ,vertex_edge)
                inner_azimuth = where((vertex_edge.cdf_factor >= 0.0) ,inner_azimuth ,vertex_azimuth)
                outer_edge = where((vertex_edge.cdf_factor >= 0.0) ,vertex_edge ,outer_edge)
                outer_azimuth = where((vertex_edge.cdf_factor >= 0.0) ,vertex_azimuth ,outer_azimuth)
            azimuth_0 = polygon.vertex_azimuths[i]
            azimuth_1 = polygon.vertex_azimuths[i + 1]
        target_projected_solid_angle += sector_projected_solid_angle
        random_numbers[0] = target_projected_solid_angle / sector_projected_solid_angle
        random_numbers[0] = clip(random_numbers[0], 0.0, 1.0)
        return sample_sector_between_edges(random_numbers, target_projected_solid_angle, inner_edge, inner_azimuth, outer_edge, outer_azimuth, azimuth_0, azimuth_1, iteration_count)


def compute_projected_solid_angle_polygon_sampling_error_arvo(polygon, random_numbers, sampled_dir):
    target_projected_solid_angle = random_numbers[0] * polygon.projected_solid_angle
    if (polygon.inner_edge_0.cdf_factor > 0.0):
        return vec2(0.0)
    else:
        outer_edge = edge_arvo_t()
        inner_edge = polygon.inner_edge_0
        inner_azimuth = polygon.vertex_azimuths[0]
        for i in range(0, MAX_POLYGON_VERTEX_COUNT - 1):
            if ((i > 1 and i + 1 == polygon.vertex_count) or (i > 0 and target_projected_solid_angle < 0.0)):
                break
            sector_projected_solid_angle = polygon.sector_projected_solid_angles[i]
            target_projected_solid_angle -= sector_projected_solid_angle
            vertex_edge = polygon.edges[i]
            vertex_azimuth = polygon.vertex_azimuths[i]
            if (i == 0):
                outer_edge = vertex_edge
                outer_azimuth = vertex_azimuth
            else:
                inner_edge = where((vertex_edge.cdf_factor >= 0.0) ,inner_edge ,vertex_edge)
                inner_azimuth = where((vertex_edge.cdf_factor >= 0.0) ,inner_azimuth ,vertex_azimuth)
                outer_edge = where((vertex_edge.cdf_factor >= 0.0) ,vertex_edge ,outer_edge)
                outer_azimuth = where((vertex_edge.cdf_factor >= 0.0) ,vertex_azimuth ,outer_azimuth)
            azimuth_0 = polygon.vertex_azimuths[i]
        target_projected_solid_angle += sector_projected_solid_angle
        sampled_azimuth = arctan2(sampled_dir[1], sampled_dir[0])
        outer_psa = get_edge_projected_solid_angle_in_sector_derivative_arvo(
            outer_edge, azimuth_0 - outer_azimuth, sampled_azimuth - outer_azimuth)[0]
        inner_psa = get_edge_projected_solid_angle_in_sector_derivative_arvo(
            inner_edge, azimuth_0 - inner_azimuth, sampled_azimuth - inner_azimuth)[0]
        sampled_projected_solid_angle = outer_psa + inner_psa
        return vec2(
            (target_projected_solid_angle - sampled_projected_solid_angle) / polygon.projected_solid_angle,
            target_projected_solid_angle - sampled_projected_solid_angle)
