#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re
from slide_utility import *
from generate_graphics import plot_polygon, get_decentral_quad, prepare_polygon, get_euler_projection, plot_axis, \
    plot_polygon_sampling, sample_sector_between_ellipses, sample_edges, solid_angle_color, solid_angle_line_color, \
    sampled_area_color, great_circle_color, sector_color, current_area_color, area_discrepancy_color, horizon_color
from polygon_clipping import clip_polygon
from polygon_sampling_related_work_ported import sample_area_polygon_turk
from polygon_sampling_ported import prepare_solid_angle_polygon_sampling, sample_solid_angle_polygon
from sampling_reference_implementation import Polygon, is_inner, ellipse_area_in_sector
from subprocess import Popen
from os import path, listdir
from scipy import optimize
from matplotlib import cm


def get_random_numbers():
    """Returns 32 2D blue noise points."""
    return np.fromfile("blue_noise_de_goes_32.dat", sep=" ").reshape((-1, 2))


def title_slide(_step):
    """Plots a title slide (just a test)."""
    draw_image("../Graphics/polygon/title_slide/title_slide.png", (0.0, 0.0), 16.0)


# These images are used by several slides so it is best to load them just once
motivation_result_images = [
    imread(path.join("../vulkan_renderer/data/experiments",
        [name for name in listdir("../vulkan_renderer/data/experiments") if re.match(r"cornell_box_%s_.*\.png" % technique, name)][0]))
    for technique in ["area_turk", "solid_angle_ours", "projected_solid_angle_ours"]
]


class RendererSlide(Slide):
    """Starts my renderer for a live demo."""

    def __init__(self, experiment=-1, show_gui=False, v_sync=True, title=None):
        super().__init__()
        self.process = None
        self.title = title
        self.args = [
            "-e%d" % experiment,
            "-gui" if show_gui else "-no_gui",
            "-v_sync" if v_sync else "-no_v_sync",
        ]
        self.notes = ["IS\nPolygonal diffuse+specular\nUnbiased @ 2 spp\nLow variance\nGood IS"]

    def animate(self):
        if self.process is not None and self.process.poll() is not None:
            self.process = None
            return "step_forward"

    def on_step(self, state, previous_state):
        # Run the renderer when we move forward through the presentation
        if state[0] > previous_state[0]:
            self.process = Popen(["../vulkan_renderer/build/vulkan_renderer"] + self.args, cwd="../vulkan_renderer")
        # Just skip this slide when moving backward
        else:
            return "step_backward"

    def draw(self):
        if self.title is not None:
            draw_title(self.title)


class MotivationSlide(Slide):

    def __init__(self, measure):
        """
        A slide to illustrate benefits of different sampling strategies.
        :param measure: "area", "solid" or "projected".
        """
        super().__init__()
        self.yaw, self.pitch = -2.284, 0.154
        self.polygon_axes = None
        self.measure = measure
        # Define the polygon
        origin = np.asarray([1.2, 2.3, 0.8])
        direction_u = np.asarray([-2.3, -0.5, -1.2])
        direction_v = np.asarray([-0.4, 0.4, 1.3])
        direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
        self.polygon = prepare_polygon([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
        # Create samples uniformly with respect to the appropriate measure
        random_numbers = get_random_numbers()
        self.samples = np.zeros((random_numbers.shape[0], 3))
        vertices = self.polygon[0]
        if self.measure == "area":
            self.step_count = 4
            fan_areas = [np.linalg.det(vertices[0:3, :]), np.linalg.det(vertices[1:4, :]), 0.0]
            fan_areas = [(fan_areas[0], fan_areas[0]), (fan_areas[1], fan_areas[0] + fan_areas[0]), (0.0, fan_areas[0] + fan_areas[0])]
            for j in range(random_numbers.shape[0]):
                self.samples[j] = sample_area_polygon_turk(vertices.shape[0], vertices, fan_areas, np.copy(random_numbers[j, :]))
            self.samples /= np.linalg.norm(self.samples, axis=1)[:, np.newaxis]
        elif self.measure == "solid":
            prepared_polygon = prepare_solid_angle_polygon_sampling(vertices.shape[0], np.vstack([vertices, vertices[0, :]]), np.asarray(3 * [0.0]))
            for j in range(random_numbers.shape[0]):
                self.samples[j] = sample_solid_angle_polygon(prepared_polygon, np.copy(random_numbers[j, :]))
        elif self.measure == "projected":
            prepared_polygon = Polygon(vertices)
            self.samples = prepared_polygon.sample(random_numbers.T).T
            self.notes = ["PSA sampling\nLight, SA\nSample cos\nMC cos cancels\nDiffuse, specular"]
        elif self.measure == "ltc":
            shading_pitch = 0.5 * np.pi
            world_to_shading_space = np.asarray([
                [np.cos(shading_pitch), -np.sin(shading_pitch), 0.0],
                [np.sin(shading_pitch), np.cos(shading_pitch), 0.0],
                [0.0, 0.0, 1.0],
            ])
            # shading_to_cosine_space = np.asarray([[0.733347774, 0.000000, -0.0798025057], [0.000000, 0.786373079, 0.000000], [0.113918357, 0.000000, 1.0]]).T
            # shading_to_cosine_space = np.asarray([[0.00990602188, 0.000000, -0.00124329492], [0.000000, 1.0, 0.000000], [0.0985978469, 0.000000, 0.000127096064]]).T
            # shading_to_cosine_space = np.asarray([[0.0105073005, 0.000000, -0.00500859041], [0.000000, 1.0, 0.000000], [0.102492839, 0.000000, 0.000652103685]]).T
            # shading_to_cosine_space = np.asarray([[0.0309120584, 0.000000, -0.0297419745], [0.000000, 1.0, 0.000000], [0.183499739, 0.000000, 0.0111321807]]).T
            shading_to_cosine_space = np.asarray([[0.117774136, 0.000000, -0.105238415], [0.000000, 1.0, 0.000000], [0.319090545, 0.000000, 0.0845179483]]).T
            world_to_cosine_space = shading_to_cosine_space @ world_to_shading_space
            prepared_polygon = Polygon(vertices @ world_to_cosine_space.T)
            self.samples = prepared_polygon.sample(random_numbers.T).T @ np.linalg.inv(world_to_cosine_space).T
            self.samples /= np.linalg.norm(self.samples, axis=-1)[..., np.newaxis]
        # Project them onto the light source
        self.light_normal = -np.cross(vertices[1, :] - vertices[0, :], vertices[2, :] - vertices[0, :])
        self.light_normal /= np.linalg.norm(self.light_normal)
        light_offset = self.light_normal @ vertices[0, :]
        self.light_samples = self.samples * light_offset / np.einsum("...i, i -> ...", self.samples, self.light_normal)[:, np.newaxis]

    def draw(self):
        title_dict = dict(area="Area sampling", solid="Solid angle sampling",
                          projected="Projected solid angle sampling (ours)", ltc="LTC importance sampling")
        draw_title(title_dict[self.measure])
        text_x = 7.6
        rendering_equation_steps = [
            r"$L_o(\omega_o)=     \int_\Omega L_i(\omega_i) f_r(\omega_i, \omega_o) n^\mathsf{T}\omega_i \,\mathrm{d}\omega_i$",
            r"$L_o(\omega_o)= L_e \int_\mathrm{\mathbb{P}}   V\,(\omega_i) f_r(\omega_i, \omega_o) n^\mathsf{T}\omega_i \,\mathrm{d}\omega_i$",
        ]
        if self.measure == "area":
            pyplot.text(text_x, 6.0, rendering_equation_steps[np.minimum(self.step, 1)])
            if self.step >= 2:
                pyplot.text(text_x, 4.0, r"$L_o(\omega_o) \approx L_e V\,(\omega_i) f_r(\omega_i, \omega_o) \, \frac{n^\mathsf{T}\omega_i}{p(\omega_i)}$")
            if self.step >= 3:
                pyplot.text(text_x, 2.7, r"$~~~~~~~~~\,= L_e V\,(\omega_i) f_r(\omega_i, \omega_o) \, \frac{n^\mathsf{T}\omega_i|n_l^\mathsf{T}\omega_i|}{\|x-y\|^2} A$")
                pyplot.text(text_x, 1.7, r"$A\,\coloneq\,$area")
        elif self.measure == "solid":
            pyplot.text(text_x, 6.0, rendering_equation_steps[1])
            pyplot.text(text_x, 4.0, r"$L_o(\omega_o) \approx L_e V\,(\omega_i) f_r(\omega_i, \omega_o) \, n^\mathsf{T}\omega_i A$")
            pyplot.text(text_x, 2.7, r"$A\,\coloneq\,$solid angle")
        elif self.measure == "projected":
            pyplot.text(text_x, 6.0, rendering_equation_steps[1])
            pyplot.text(text_x, 4.0, r"$L_o(\omega_o) \approx L_e V\,(\omega_i) f_r(\omega_i, \omega_o) A$")
            pyplot.text(text_x, 2.7, r"$A\,\coloneq\,$projected solid angle")
        elif self.measure == "ltc":
            pyplot.text(text_x, 6.0, rendering_equation_steps[1])
            pyplot.text(text_x, 4.0, r"$L_o(\omega_o) \approx V\,(\omega_i)A$")
            pyplot.text(text_x, 2.7, r"$A\,\coloneq\,$unshadowed shading")
        slide_axes = pyplot.gca()
        self.polygon_axes = slide_axes.inset_axes((0.3, 0.7, 7.1, 7.1), transform=slide_axes.transData,
                                                  aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.polygon_axes)
        self.draw_polygon()

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.draw_polygon()
            return "redraw_canvas"

    def draw_polygon(self):
        self.polygon_axes.clear()
        self.polygon_axes.set_xlim(-1.05, 2.8)
        self.polygon_axes.set_ylim(-0.85, 3.0)
        self.polygon_axes.set_xticks([])
        self.polygon_axes.set_yticks([])
        pyplot.sca(self.polygon_axes)
        pyplot.scatter([0.0], [0.0], color="k", s=4.0)
        elements_dict = dict(
            area=(("polygon", "hemisphere") if self.step == 0 else ("polygon", "hemisphere", "projected_polygon")),
            solid=("polygon", "hemisphere", "projected_polygon"),
            projected=("polygon_red", "clipped_polygon", "hemisphere", "projected_polygon"),
            ltc=("polygon_red", "clipped_polygon", "hemisphere", "projected_polygon"),
        )
        projection = get_euler_projection(self.yaw, self.pitch)
        plot_polygon(self.polygon, projection, elements_dict[self.measure])
        # Plot normal vectors
        vertices = self.polygon[0]
        origins = [np.asarray(3 * [0.0]), vertices.mean(axis=0)]
        normals = [np.asarray([0.0, 0.0, 1.0]), self.light_normal]
        normal_count = 2 if self.measure == "area" and self.step >= 3 else 1
        for label, origin, normal in list(zip(("$n$", "$n_l$"), origins, normals))[0:normal_count]:
            transformed_origin = projection.transform(origin)
            transformed_normal = projection.transform(normal)
            length = 0.88
            pyplot.arrow(transformed_origin[0], transformed_origin[1], length * transformed_normal[0], length * transformed_normal[1],
                         head_width=0.08, facecolor="w")
            label_position = transformed_origin + transformed_normal * 0.6
            pyplot.text(label_position[0], label_position[1], label, ha="left", va="bottom")
        # Label the solid angle of the polygon
        if self.measure == "area" and self.step >= 1:
            polygon_label_position = projection.transform(origins[1] / np.linalg.norm(origins[1]))
            pyplot.text(polygon_label_position[0], polygon_label_position[1], r"$\mathrm{\mathbb{P}}$", ha="center", va="center")
        # Draw the samples
        samples_list = list()
        if self.measure == "area":
            if self.step >= 3:
                samples_list = [self.light_samples]
        else:
            samples_list = [self.samples]
        for samples in samples_list:
            transformed_samples = projection.transform(samples)
            pyplot.scatter(transformed_samples[:, 0], transformed_samples[:, 1], s=1.0, c="k")


class FastMotivationSlide(Slide):

    def __init__(self):
        """
        A slide to illustrate projected solid angle sampling quickly.
        """
        super().__init__()
        self.yaw, self.pitch = -2.284, 0.154
        self.polygon_axes = None
        # Define the polygon
        origin = np.asarray([1.2, 2.3, 0.8])
        direction_u = np.asarray([-2.3, -0.5, -1.2])
        direction_v = np.asarray([-0.4, 0.4, 1.3])
        direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
        self.polygon = prepare_polygon([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
        # Create samples uniformly with respect to the appropriate measure
        random_numbers = get_random_numbers()
        self.samples = np.zeros((random_numbers.shape[0], 3))
        vertices = self.polygon[0]
        prepared_polygon = Polygon(vertices)
        self.samples = prepared_polygon.sample(random_numbers.T).T

    def draw(self):
        draw_title("Projected solid angle sampling")
        pyplot.text(5.9, 1.0, r"$p(\omega_i)\propto n\cdot\omega_i$", ha="center", va="bottom", size=25.0)
        slide_axes = pyplot.gca()
        self.polygon_axes = slide_axes.inset_axes((0.3, 0.7, 16.0, 7.5), transform=slide_axes.transData,
                                                  aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.polygon_axes)
        self.draw_polygon()

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.draw_polygon()
            return "redraw_canvas"

    def draw_polygon(self):
        self.polygon_axes.clear()
        self.polygon_axes.set_xlim(-1.05, 2.8)
        self.polygon_axes.set_ylim(-0.85, 2.0)
        self.polygon_axes.set_xticks([])
        self.polygon_axes.set_yticks([])
        pyplot.sca(self.polygon_axes)
        pyplot.scatter([0.0], [0.0], color="k", s=4.0)
        elements = ("polygon_red", "clipped_polygon", "hemisphere", "projected_polygon")
        projection = get_euler_projection(self.yaw, self.pitch)
        plot_polygon(self.polygon, projection, elements)
        # Plot normal vectors
        vertices = self.polygon[0]
        transformed_normal = projection.transform(np.asarray([0.0, 0.0, 1.0]))
        length = 0.88
        pyplot.arrow(0.0, 0.0, length * transformed_normal[0], length * transformed_normal[1],
                     head_width=0.08, facecolor="w")
        label_position = transformed_normal * 0.6
        pyplot.text(label_position[0], label_position[1], r"$n$", ha="left", va="bottom")
        # Draw the samples
        samples_list = list()
        transformed_samples = projection.transform(self.samples)
        pyplot.scatter(transformed_samples[:, 0], transformed_samples[:, 1], s=1.0, c="k")


class MotivationResultSlide(ImageComparisonSlide):
    """Shows Cornell box renderings."""

    def __init__(self, image_count=3):
        super().__init__()
        self.images = motivation_result_images[0:image_count]
        self.labels = ["Area", "Solid angle", "Projected solid angle"]
        self.left_bottom = 0.35, 1.2
        self.image_width = 5.0
        self.offset = 5.1, 0.0
        if image_count == 3:
            self.notes = ["Good everywhere\nExcept penumbra\nRandom hit/miss"]

    def draw(self):
        super().draw()
        draw_title(("Results" if len(self.images) > 1 else "Result") + " at 1 sample per pixel")


class PolygonSlide(Slide):
    """
    Base class for slides that use a particular polygon as example and display
    it in a particular way.
    """

    def __init__(self, show_axes=False):
        super().__init__()
        self.polygon = get_decentral_quad()
        self.prepared_polygon = Polygon(self.polygon[0])
        self.slide_axes = None
        self.polygon_axes = None
        self.yaw, self.pitch = -3.0, 0.154
        self.projection = None
        self.update_projection(2)
        self.show_axes = show_axes
        self.sector_count = self.prepared_polygon.vertex_count - 1

    def create_polygon_axes(self):
        """
        Creates axes for displaying the polygon and adds them to the figure.
        """
        self.slide_axes = pyplot.gca()
        self.polygon_axes = self.slide_axes.inset_axes((3.0, 0.3, 13.0, 7.8), transform=self.slide_axes.transData, aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.polygon_axes)

    def prepare_polygon_axes(self, polygon_zoom=1):
        """
        Clears self.polygon_axes and prepares it for displaying the polygon.
        :param polygon_zoom: Defines the axis limits. 0 if the whole polygon is
            shown, 1 if the sphere is shown, 2 for transition to 3, 3 if only a
            single sector is shown.
        """
        self.polygon_axes.clear()
        if polygon_zoom == 0:
            self.polygon_axes.set_xlim(-1.3, smooth_lerp(2.8, 2.3, self.step_lerp))
            self.polygon_axes.set_ylim(smooth_lerp(-0.85, -1.1, self.step_lerp), smooth_lerp(3.0, 1.1, self.step_lerp))
        elif polygon_zoom == 1:
            self.polygon_axes.set_xlim(-1.3, 2.3)
            self.polygon_axes.set_ylim(-1.1, 1.1)
        elif polygon_zoom == 2:
            self.polygon_axes.set_xlim(smooth_lerp(-1.3, -0.7, self.step_lerp), smooth_lerp(2.3, 0.95, self.step_lerp))
            self.polygon_axes.set_ylim(-1.1, smooth_lerp(1.1, 0.1, self.step_lerp))
        elif polygon_zoom == 3:
            self.polygon_axes.set_xlim(-0.7, 0.95)
            self.polygon_axes.set_ylim(-1.1, 0.1)
        self.polygon_axes.set_xticks([])
        self.polygon_axes.set_yticks([])
        pyplot.sca(self.polygon_axes)

    def update_projection(self, projection_step):
        """
        Sets self.projection, either based on the current yaw and pitch or to a
        fixed top view (or lerped in between).
        :param projection_step: <=0 for free, 1 for lerp, >=2 for fixed
        :return: The used lerp factor between dynamic and fixed.
        """
        if projection_step <= 0:
            lerp_factor = 0.0
        elif projection_step == 1:
            lerp_factor = self.step_lerp
        else:
            lerp_factor = 1.0
        self.projection = get_euler_projection(smooth_lerp(self.yaw, -np.pi, lerp_factor), smooth_lerp(self.pitch, 0.5 * np.pi, lerp_factor))
        return lerp_factor

    def planar_projection(self, points):
        """
        Appends a 0 z-coordinate to the given 2D points and applies the
        projection transform.
        """
        return self.projection.transform(np.concatenate([points, 0.0 * points[..., 0:1]], axis=-1))

    def get_sector_ellipses(self, sector_index):
        """
        :param sector_index: Index of a sector of the held polygon.
        :return: A pair inner, outer providing ellipses for this sector as 2D
            vectors.
        """
        inner = self.prepared_polygon.inner_ellipse_0
        outer = self.prepared_polygon.ellipses[0]
        for j in range(1, sector_index + 1):
            ellipse = self.prepared_polygon.ellipses[j]
            if is_inner(ellipse):
                inner = ellipse
            else:
                outer = ellipse
        return inner, outer

    def get_sector_outline(self, sector_index, sector_begin=None, sector_end=None):
        """
        :param sector_index: The index of a sector in the polygon.
        :param sector_begin: Optional override for the beginning of the sector.
        :param sector_end: Optional override for the end of the sector.
        :return: An array of shape (sample_count, 3) providing the outline of
            the sector between the ellipses in the xy-plane.
        """
        inner, outer = self.get_sector_ellipses(sector_index)
        if sector_begin is None:
            sector_begin = self.prepared_polygon.vertices[sector_index]
        if sector_end is None:
            sector_end = self.prepared_polygon.vertices[sector_index + 1]
        edges = list()
        lerp_factor = np.linspace(0.0, 1.0, 129)[:, np.newaxis]
        for ellipse in [inner, outer]:
            edge = lerp_factor * sector_begin[np.newaxis, :] + (1.0 - lerp_factor) * sector_end[np.newaxis, :]
            matrix = np.eye(2) + np.outer(ellipse, ellipse)
            edge /= np.sqrt(np.einsum("ij, jk, ik -> i", edge, matrix, edge)[:, np.newaxis])
            edges.append(edge)
        outline_2d = np.concatenate([edges[0], edges[1][::-1, :]])
        return np.hstack([outline_2d, 0.0 * outline_2d[:, 0:1]])


class GeometricConstructionsSlide(PolygonSlide):
    """Explains the structure of the projected solid angle of a polygon."""

    def __init__(self, forced_step=-1, show_axes=True, initial_sector=-1, show_bullets=True):
        """
        Prepares the polygon. Pass a step index if you want to show this step
        only. You can also turn 3D axes and certain bullet texts on or off.
        """
        super().__init__(show_axes)
        self.forced_step = forced_step
        if self.forced_step < 0:
            self.step_count = 6
            self.step_durations = [2.0, 0.0, 2.0]
        if self.forced_step == 6:
            self.notes = ["Consider PSA\nSA proj. along normal\nSample uniformly"]
        if self.forced_step == 5:
            self.notes = ["Cut apart\nLines 0->vert\n2 edges\nArcs of ellipses\nSimpler"]
        self.sector = initial_sector
        self.show_bullets = show_bullets

    def on_click(self, position, button):
        return self.on_mouse_move(position, (0, 0), {button})

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons and self.get_step() <= 2:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.draw_polygon()
            return "redraw_canvas"
        if self.get_step() == 5 and "lmb" in held_buttons:
            position_data = self.polygon_axes.transData.inverted().transform(position)
            position_world = self.projection.inverse_transform([position_data[0], position_data[1], 0.0])
            direction = np.asarray([position_world[0], position_world[1]])
            sector = -1
            for j in range(self.sector_count):
                begin, end = self.prepared_polygon.vertices[j], self.prepared_polygon.vertices[j + 1]
                if np.linalg.det(np.vstack([direction, begin])) <= 0.0 and np.linalg.det(np.vstack([direction, end])) > 0.0:
                    sector = j
                    break
            if sector != self.sector:
                self.sector = sector
                self.draw_polygon()
                return "redraw_canvas"

    def draw(self):
        draw_title("Projected solid angle")
        self.create_polygon_axes()
        self.draw_polygon()
        pyplot.sca(self.slide_axes)
        if self.get_step() == 4:
            pyplot.text(1.0, 4.8, r"$u_j\,\coloneq\, \frac{n_{j,\mathrm{xy}}}{n_{j,\mathrm{z}}}\in\mathrm{\mathbb{R}}^2$")
            pyplot.text(1.0, 3.7, r"$C_j\,\coloneq\, I + u_ju_j^\mathsf{T}$")
            pyplot.text(1.0, 2.0, r"$q\in\mathrm{\mathbb{R}}^2$ in ellipse" + "\n" + "$\Leftrightarrow \quad q^\mathsf{T}C_jq\leq 1$")
        if self.get_step() == 5 and self.show_bullets:
            pyplot.text(0.7, 1.7,r"Cut through vertices")
            pyplot.text(0.7, 0.9, r"2 ellipses per sector")

    def animate(self):
        if self.lerp_ongoing:
            self.draw_polygon()
            return "redraw_canvas"

    def get_step(self):
        """Returns either self.step or self.forced_step."""
        return self.step if self.forced_step < 0 else self.forced_step

    def draw_polygon(self):
        self.prepare_polygon_axes(0 if self.get_step() < 1 else 1)
        circles = [(3, 4), (3, 0), (2, 0), (1, 0)]
        step_elements = [
            ("polygon_red", "clipped_polygon", "hemisphere", "projected_polygon"),
            ("hemisphere", "projected_polygon"),
            ("hemisphere", "projected_polygon", "great_circle_4", "unlabeled_normals" if self.lerp_ongoing else "simple_label_normals"),
            ("hemisphere", "projected_polygon", "great_circle_4", "unlabeled_normals"),
            ("horizon", "projected_polygon", "great_circle_4", "ellipse_normals"),
            ["horizon", "projected_polygon", "sectors"] + (["great_circle_%d" % circles[self.sector][0], "great_circle_%d" % circles[self.sector][1]] if self.sector >= 0 else []),
            ("horizon", "projected_polygon"),
        ]
        projection_lerp = self.update_projection(self.get_step() - 1)
        y_offset = smooth_lerp(0.5, 0.15, projection_lerp)
        plot_polygon(self.polygon, self.projection, step_elements[self.get_step() + (1 if self.get_step() == 0 and self.step_lerp > 0.0 else 0)])
        if self.show_axes:
            for j in range(3):
                plot_axis(self.projection, np.asarray([-2.0, -0.6, -0.5]), j, False, [0.15, y_offset, 0.15][j])
        # Mark the area inside the highlighted sector
        if self.get_step() == 5 and self.sector >= 0:
            outline = self.get_sector_outline(self.sector)
            outline = self.projection.transform(outline)
            pyplot.fill(outline[:, 0], outline[:, 1], color=sampled_area_color)


class AreaComputationSlide(PolygonSlide):
    """Visualizes computation of the area within a sector."""

    def __init__(self):
        super().__init__()
        self.ellipse = self.prepared_polygon.ellipses[2, :]
        self.sector = self.prepared_polygon.vertices[2:4, :]
        self.step_count = 4
        self.step_durations = [2.0, 0.0, 2.0]
        self.update_projection(2)
        # Prepare to plot the ellipse
        phase = np.linspace(-np.pi, np.pi, 513)
        circle = np.vstack([np.cos(phase), np.sin(phase)])
        self.matrix = np.eye(2) + np.outer(self.ellipse, self.ellipse)
        eigenvalues, eigenvectors = np.linalg.eigh(self.matrix)
        self.root = eigenvectors @ np.diag(np.sqrt(eigenvalues)) @ eigenvectors.T
        self.ellipse_samples = (np.linalg.inv(self.root) @ circle).T
        # Scale the sector vectors onto the ellipse
        scaling = 1.0 / np.sqrt(np.einsum("ij, jk, ik -> i", self.sector, self.matrix, self.sector))
        self.sector *= scaling[:, np.newaxis]
        # Create an outline for the sector
        sector_lines = np.vstack([self.sector[0], [0.0, 0.0], self.sector[1]])
        lerp_factor = np.linspace(0.0, 1.0, 258)[1:257]
        ellipse_arc = lerp_factor[:, np.newaxis] * self.sector[0:1, :] + (1.0 - lerp_factor[:, np.newaxis]) * self.sector[1:2, :]
        scaling = 1.0 / np.sqrt(np.einsum("ij, jk, ik -> i", ellipse_arc, self.matrix, ellipse_arc))
        ellipse_arc *= scaling[:, np.newaxis]
        self.sector_outline = np.vstack([sector_lines, ellipse_arc])

    def projection_transform(self, points):
        """
        Applies the plane transform followed by the projection transform to
        points in the xy-plane (shape (..., 2)).
        """
        if self.step == 0:
            transform_lerp = self.step_lerp
        elif self.step == 1:
            transform_lerp = 1.0
        elif self.step == 2:
            transform_lerp = 1.0 - self.step_lerp
        else:
            transform_lerp = 0.0
        transform = smooth_lerp(np.eye(2), self.root, transform_lerp)
        return -np.einsum("ij, ...j -> ...i", transform, points)

    def animate(self):
        if self.lerp_ongoing:
            self.draw_area()
            return "redraw_canvas"

    def draw(self):
        draw_title("Projected solid angle")
        self.create_polygon_axes()
        self.draw_area()
        pyplot.sca(self.slide_axes)
        if self.step >= 2:
            texts = [
                r"$\alpha = \mathrm{atan2}\binom{-|(s_0, s_1)|} {\frac{1}{|C_j|} s_0^\mathsf{T} C_j s_1}$",
                r"$A = \frac{1}{2|C_j|} \mathrm{atan2}\binom{-|(s_0, s_1)|} {\frac{1}{|C_j|} s_0^\mathsf{T} C_j s_1}$",
            ]
            pyplot.text(15.0, 1.7, texts[self.step - 2], ha="right", va="center")

    def draw_area(self):
        self.prepare_polygon_axes()
        # Fill and plot the sector
        sector_outline = self.projection_transform(self.sector_outline)
        pyplot.fill(sector_outline[:, 0], sector_outline[:, 1], color=solid_angle_color, linewidth=0.0)
        sector = self.projection_transform(self.sector)
        for j in range(2):
            pyplot.arrow(0.0, 0.0, sector[j, 0], sector[j, 1], head_width=0.04, facecolor="w", zorder=100.0)
        # Label the sector boundaries
        if self.step in {0, 3} and self.step_lerp == 0.0:
            for j in range(2):
                boundary = sector[j, :]
                offset = [-1.0, 1.0][j] * np.asarray([-boundary[1], boundary[0]])
                offset *= 0.08 / np.linalg.norm(offset)
                label_position = boundary * 0.6 + offset
                pyplot.text(label_position[0], label_position[1], r"$s_%d$" % j, ha="center", va="center")
            # Label the ellipse
            ellipse_label_position = np.asarray([1.0, 0.7])
            ellipse_label_position /= np.sqrt(ellipse_label_position.T @ self.matrix @ ellipse_label_position)
            ellipse_label_position = self.projection_transform(ellipse_label_position)
            pyplot.text(ellipse_label_position[0], ellipse_label_position[1] - 0.05, r"$C_j$", ha="center", va="top")
        # Label the area
        if self.step == 3:
            fill_center = sector_outline.mean(axis=0) * 0.6
            pyplot.text(fill_center[0], fill_center[1], r"$A$", ha="center", va="center")
        # Draw an angle
        if self.step == 2 and self.step_lerp == 0.0:
            angle = sector_outline[3:, :]
            angle = angle * 0.3 / np.linalg.norm(angle, axis=1)[:, np.newaxis]
            pyplot.plot(angle[:, 0], angle[:, 1], color="k")
            angle_center = np.mean(angle, axis=0) * 0.7
            pyplot.text(angle_center[0], angle_center[1], r"$\alpha$", ha="center", va="center")
        # Plot the ellipse
        transformed_ellipse = self.projection_transform(self.ellipse_samples)
        pyplot.plot(transformed_ellipse[:, 0], transformed_ellipse[:, 1], color=great_circle_color)


def draw_primary_sample_space_rectangle(width_scale, color):
    """
    Draws a rectangle visualizing primary sample space at a standard location.
    :param width_scale: By how much the width is scaled.
    :param color: The color of the fill of the rectangle.
    :return: (left, top, right, bottom) for the rectangle in data coordinates.
    """
    left, top = 1.0, 7.0
    width, height = width_scale * 2.8, 2.8
    right, bottom = left + width, top - height
    rectangle = np.asarray([(left, top), (right, top), (right, bottom), (left, bottom)])
    pyplot.fill(rectangle[:, 0], rectangle[:, 1], linewidth=1.0, edgecolor="k", color=color)
    return left, top, right, bottom


class AzimuthSamplingSlide(PolygonSlide):
    """Shows how an azimuth should be chosen to implement sampling correctly"""

    def __init__(self, short_talk=False):
        super().__init__()
        self.sampled_dir = self.prepared_polygon.vertices[0, :]
        self.step_count = 1 if short_talk else 3
        # Projected sector outlines for all sectors
        self.outlines = [self.projection.transform(self.get_sector_outline(j)) for j in range(self.sector_count)]
        if short_talk:
            self.notes = ["Given rand()\nRatio left/right equal\nRatio for dir: Easy\nDir for ratio: Hard"]

    def draw(self):
        draw_title("Sampling the projected solid angle")
        self.create_polygon_axes()
        area, sector_index = self.draw_polygon()
        # Plot primary sample space
        pyplot.sca(self.slide_axes)
        draw_primary_sample_space_rectangle(1.0, solid_angle_color)
        random_number_0 = area / self.prepared_polygon.projected_solid_angle
        left, top, right, bottom = draw_primary_sample_space_rectangle(random_number_0, sampled_area_color)
        pyplot.text(right, bottom - 0.1, r"$\xi_0$", ha="center", va="top")
        # Plot the sampled point
        if self.step >= 1 and sector_index >= 0:
            inner, outer = self.get_sector_ellipses(sector_index)
            length = self.sampled_dir @ self.sampled_dir
            inner_factor = length + (self.sampled_dir @ inner) ** 2
            outer_factor = length + (self.sampled_dir @ outer) ** 2
            random_number_1 = (inner_factor * outer_factor - outer_factor) / (inner_factor - outer_factor)
            sample_y = lerp(bottom, top, random_number_1)
            pyplot.plot([left, right], [sample_y, sample_y], color="k", linewidth=1.0)
            pyplot.scatter([right], [sample_y], color="w", edgecolors="k", zorder=1000.0)
            pyplot.text(left - 0.1, sample_y, r"$\xi_1$", ha="right", va="center")
        # Show why sampling is difficult
        if self.step >= 2:
            pyplot.text(9.8, 0.8, r"$a\,\mathrm{atan}(bx) + c\,\mathrm{atan}(dx)=\xi_0$")

    def on_click(self, position, button):
        """Updates the sampled direction."""
        if "lmb" == button:
            position_data = self.polygon_axes.transData.inverted().transform(position)
            self.sampled_dir = self.projection.inverse_transform([position_data[0], position_data[1], 0.0])[0:2]
            return "draw"

    def on_mouse_move(self, position, motion, held_buttons):
        """Updates the sampled direction."""
        if "lmb" in held_buttons:
            position_data = self.polygon_axes.transData.inverted().transform(position)
            self.sampled_dir = self.projection.inverse_transform([position_data[0], position_data[1], 0.0])[0:2]
            return "draw"

    def draw_polygon(self):
        """
        Draws the polygon with the area counterclockwise of the sampled
        direction highlighted.
        :return: Size of the highlighted area and the index of the sampled
            sector (or -1).
        """
        self.prepare_polygon_axes()
        normalized_dir = self.sampled_dir / np.linalg.norm(self.sampled_dir)
        # Figure out if the sampled direction covers the polygon
        polygon_covered = np.linalg.det(np.vstack([normalized_dir, self.prepared_polygon.vertices[0, :]])) < 0.0
        polygon_covered &= np.linalg.det(np.vstack([normalized_dir, self.prepared_polygon.vertices[self.sector_count, :]])) > 0.0
        # Mark the sampled direction
        if polygon_covered:
            sampled_dir = self.planar_projection(normalized_dir)
            pyplot.plot([0.0, sampled_dir[0]], [0.0, sampled_dir[1]], color="k", zorder=100.0)
        # Highlight sectors clockwise of the sample
        area = 0.0
        sector_index = -1
        for j in range(self.sector_count if polygon_covered else 0):
            clockwise_1 = np.linalg.det(np.vstack([normalized_dir, self.prepared_polygon.vertices[j + 1, :]])) < 0.0
            if clockwise_1:
                pyplot.fill(self.outlines[j][:, 0], self.outlines[j][:, 1], color=sampled_area_color, linewidth=0.0)
                area += self.prepared_polygon.sector_projected_solid_angles[j]
            clockwise_0 = np.linalg.det(np.vstack([normalized_dir, self.prepared_polygon.vertices[j, :]])) < 0.0
            if clockwise_0 and not clockwise_1:
                outline = self.get_sector_outline(j, None, normalized_dir)
                outline = self.projection.transform(outline)
                pyplot.fill(outline[:, 0], outline[:, 1], color=sampled_area_color, linewidth=0.0)
                inner, outer = self.get_sector_ellipses(j)
                area -= ellipse_area_in_sector(inner, self.prepared_polygon.vertices[j, :], normalized_dir)
                area += ellipse_area_in_sector(outer, self.prepared_polygon.vertices[j, :], normalized_dir)
                sector_index = j
        # Plot the polygon itself
        elements = ("horizon", "projected_polygon", "sectors")
        plot_polygon(self.polygon, self.projection, elements)
        # Clamp the sampled point into the polygon
        length = self.sampled_dir @ self.sampled_dir
        inner, outer = self.get_sector_ellipses(sector_index)
        inner_factor = length + (self.sampled_dir @ inner) ** 2
        outer_factor = length + (self.sampled_dir @ outer) ** 2
        if inner_factor < 1.0:
            self.sampled_dir /= np.sqrt(inner_factor)
        if outer_factor > 1.0:
            self.sampled_dir /= np.sqrt(outer_factor)
        # Plot the sampled point
        if polygon_covered and self.step >= 1:
            sampled_dir = self.planar_projection(self.sampled_dir)
            pyplot.scatter([sampled_dir[0]], [sampled_dir[1]], zorder=1000.0, color="w", edgecolors="k")
        return area, sector_index


class InitializationSlide(PolygonSlide):
    """Visualizes the initialization strategy."""

    def __init__(self):
        super().__init__()
        # Define the quads
        self.quads = np.zeros((self.sector_count, 2, 2, 2, 2))
        self.half_vectors = np.zeros((self.sector_count, 2))
        self.step_count = 6
        for i in range(self.sector_count):
            begin = self.prepared_polygon.vertices[i, :]
            end = self.prepared_polygon.vertices[i + 1, :]
            begin /= np.linalg.norm(begin)
            end /= np.linalg.norm(end)
            middle = begin + end
            middle /= np.linalg.norm(middle)
            self.half_vectors[i, :] = middle
            inner, outer = self.get_sector_ellipses(i)
            for j, sector in enumerate([(begin, middle), (middle, end)]):
                for k, ellipse in enumerate([inner, outer]):
                    for l, boundary in enumerate(sector):
                        self.quads[i, j, k, l, :] = boundary / np.sqrt(1.0 + (ellipse @ boundary) ** 2)
        # Sample the quads
        self.random_numbers = get_random_numbers()
        self.quad_samples = self.prepared_polygon.sample(self.random_numbers.T, False, 0).T

    def draw(self):
        draw_title("Initialization")
        draw_bullets([
            "Halve sectors",
            "Form quads",
            "Sample quads",
        ][0:[0, 1, 1, 2, 3, 3][self.step]], 1.0, 5.0)
        self.create_polygon_axes()
        self.draw_polygon()

    def draw_polygon(self):
        self.prepare_polygon_axes()
        # Plot the polygon itself
        elements = ("horizon", "projected_polygon_thin", "sectors")
        plot_polygon(self.polygon, self.projection, elements)
        # Plot the half vectors
        if self.step >= 1:
            half_vectors = np.zeros((self.sector_count * 2, 2))
            half_vectors[1::2, :] = self.planar_projection(self.half_vectors)[:, 0:2]
            pyplot.plot(half_vectors[:, 0], half_vectors[:, 1], color=sector_color, linewidth=1.0)
        # Plot the vertices of the quads
        quads = self.planar_projection(self.quads)[..., 0:2]
        if self.step >= 2:
            pyplot.scatter(quads[..., 0].flat, quads[..., 1].flat, color="w", edgecolors="k", zorder=10000.0, s=14.0)
        # Plot the quads
        if self.step >= 3:
            flat_quads = quads.reshape((-1, 2, 2, 2))
            for i in range(flat_quads.shape[0]):
                quad = flat_quads[i, ...]
                outline = np.concatenate([quad[0, :, :], quad[1, ::-1, :]], axis=0)
                pyplot.fill(outline[:, 0], outline[:, 1], fc=solid_angle_color, ec="k", zorder=1000.0)
        # Plot the samples
        if self.step == 4:
            quad_samples = self.projection.transform(self.quad_samples)
            pyplot.scatter(quad_samples[:, 0], quad_samples[:, 1], s=1.0, color="k", zorder=2000.0)
        # Plot the samples as lines
        if self.step >= 5:
            quad_samples = self.projection.transform(self.quad_samples)[:, 0:2]
            quad_samples /= np.linalg.norm(quad_samples, axis=-1)[..., np.newaxis]
            quad_sample_lines = np.stack([np.zeros_like(quad_samples), quad_samples], axis=0)
            pyplot.plot(quad_sample_lines[:, :, 0], quad_sample_lines[:, :, 1], color="k", linewidth=0.4, zorder=2000.0)


class IterationSlide(PolygonSlide):
    """Visualizes the iterative procedure."""

    def __init__(self, short_talk=False):
        super().__init__()
        self.projected_polygon = [polygon * np.asarray([[-1.0, -1.0, 1.0]]) for polygon in self.polygon]
        self.step_count = 8
        self.step_durations = [2.0]
        self.sector_index = 2
        self.random_number = 0.7
        target_area = self.random_number * self.prepared_polygon.projected_solid_angle
        self.sector_random_number = target_area - sum(self.prepared_polygon.sector_projected_solid_angles[0:self.sector_index])
        self.sector_random_number /= self.prepared_polygon.sector_projected_solid_angles[self.sector_index]
        self.current_dir = np.zeros((2,))
        self.initialization_random_number = 0.0
        self.set_current_dir([-0.6, 0.5])
        self.sampled_dir = self.prepared_polygon.sample(np.asarray([[self.random_number], [1.0]]))[0:2, 0]
        self.sector_outline = self.get_sector_outline(self.sector_index)
        self.short_talk = short_talk
        if self.short_talk:
            self.notes = [
                "Iterative procedure\n1 sector",
                "->",
                "Dir from init or prev\nSkipping init\nArea known",
                "Different",
                "Discrepancy",
                "Intersection, tangents",
                "Between dirs and tangents\nEllipses replaced",
                "Closer -> better\nCubic\n2 iterations\nFast, GPU-friendly",
            ]

    def on_click(self, position, button):
        """Updates the sampled direction."""
        if "lmb" == button:
            if self.set_current_dir(self.planar_projection(self.polygon_axes.transData.inverted().transform(position))):
                self.draw_polygon()
                return "draw"

    def on_mouse_move(self, position, motion, held_buttons):
        """Updates the sampled direction."""
        if "lmb" in held_buttons:
            if self.set_current_dir(self.planar_projection(self.polygon_axes.transData.inverted().transform(position))):
                self.draw_polygon()
                return "draw"

    def set_current_dir(self, new_dir):
        """
        Sets the current direction unless the given direction is not in the
        sector. Also updates related attributes.
        :return: True if an update occurred.
        """
        new_dir = np.asarray([new_dir[0], new_dir[1]])
        # Figure out if the direction is in the sector
        sector_covered = np.linalg.det(np.vstack([new_dir, self.prepared_polygon.vertices[self.sector_index, :]])) < 0.0
        sector_covered &= np.linalg.det(np.vstack([new_dir, self.prepared_polygon.vertices[self.sector_index + 1, :]])) > 0.0
        if not sector_covered:
            return False
        self.current_dir = new_dir
        # Figure out what random number for the initialization gives the
        # desired direction
        inner, outer = self.get_sector_ellipses(self.sector_index)

        def compare_initialization(random_number):
            target_area = random_number * self.prepared_polygon.sector_projected_solid_angles[self.sector_index]
            sampled_dir = sample_sector_between_ellipses(
                (random_number, 0.0), target_area, inner, outer, self.prepared_polygon.vertices[self.sector_index, :],
                self.prepared_polygon.vertices[self.sector_index + 1, :], 0)
            return np.linalg.det(np.vstack([sampled_dir, self.current_dir]))
        self.initialization_random_number = optimize.bisect(compare_initialization, 0.0, 1.0, xtol=1.0e-3, rtol=1.0e-4)
        return True

    def animate(self):
        if self.lerp_ongoing:
            self.draw_polygon()
            return "redraw_canvas"

    def draw(self):
        draw_title("Iteration")
        self.create_polygon_axes()
        self.draw_polygon()
        pyplot.sca(self.slide_axes)
        # Draw primary sample space
        draw_primary_sample_space_rectangle(1.0, solid_angle_color)
        current_area = sum([self.prepared_polygon.sector_projected_solid_angles[j] for j in range(self.sector_index)])
        inner, outer = self.get_sector_ellipses(self.sector_index)
        current_area += ellipse_area_in_sector(outer, self.prepared_polygon.vertices[self.sector_index, :], self.current_dir)
        current_area -= ellipse_area_in_sector(inner, self.prepared_polygon.vertices[self.sector_index, :], self.current_dir)
        current_random_number = current_area / self.prepared_polygon.projected_solid_angle
        if self.step >= 2:
            top, current_right, bottom = draw_primary_sample_space_rectangle(current_random_number, current_area_color if self.step <= 3 else area_discrepancy_color)[1:]
            if self.step >= 3:
                sampled_right = draw_primary_sample_space_rectangle(self.random_number, sampled_area_color)[2]
                pyplot.text(sampled_right, bottom - 0.1, r"$\xi_0$", ha="center", va="top")
                if self.step >= 4:
                    pyplot.text(0.5 * (sampled_right + current_right), 0.5 * (bottom + top), r"$A_\delta$", ha="center", va="center", fontsize=10.0)
            if self.short_talk:
                pyplot.text(0.7, 1.7, "Good initialization")
        # Explain the math
        if self.step >= self.step_count - 1:
            if self.short_talk:
                pyplot.text(0.7, 1.0, "+ 2 iterations")
            else:
                pyplot.text(4.5, 8.4, r"$T:=\binom{-w_{n,\mathrm{y}}}{w_{n,\mathrm{x}}}w_n^\mathsf{T}(C_i-C_o)-2A_\delta C_iw_nw_n^\mathsf{T}C_o\in\mathrm{\mathbb{R}}^{2\times2}$", ha="left", va="top")
                pyplot.text(4.5, 7.5, r"$w_{n+1}^\mathsf{T}T\,w_{n+1}=0$", ha="left", va="top")

    def draw_polygon(self):
        self.prepare_polygon_axes(2 if self.step == 0 else 3)
        # Plot the polygon
        plot_polygon(self.polygon, self.projection, ("projected_polygon", ))
        # Illustrate the sampling procedure
        if self.step == 0:
            elements = ("ellipses", "sector")
        elif self.step == 1:
            elements = ("ellipses", "sector")
        elif self.step <= 3:
            elements = ("ellipses", "sector", "current_direction", "current_area")
        elif self.step <= 4:
            elements = ("ellipses", "sector", "current_direction")
        elif self.step <= 5:
            elements = ("ellipses", "sector", "current_direction", "tangents", "tangent_points")
        else:
            elements = ("ellipses", "sector", "current_direction", "tangents", "tangent_points", "tangent_area_red", "next_direction")
        plot_polygon_sampling(self.projected_polygon, elements, 0, self.sector_index,
                              np.asarray([self.sector_random_number, 1.0]), self.initialization_random_number, -0.5 * np.pi)
        if "current_area" in elements:
            for j in range(self.sector_index):
                outline = self.projection.transform(self.get_sector_outline(j))
                pyplot.fill(outline[:, 0], outline[:, 1], color=current_area_color, linewidth=0.0)


class SampleMappingSlide(Slide):
    """Visualizes the mapping from primary sample space to samples."""

    def __init__(self, title="Sample mapping"):
        super().__init__()
        self.title = title
        self.vertices = get_decentral_quad()[0]
        self.vertices /= np.linalg.norm(self.vertices, axis=1)[:, np.newaxis]
        self.clipped_vertices = np.copy(self.vertices)
        self.prepared_polygon = Polygon(self.vertices)
        self.set_vertices(self.vertices)
        self.random_numbers = get_random_numbers()
        self.pss_axes = self.polygon_axes = None
        # Prepare a grid
        self.resolution = (12, 12, 16)
        grid_0, grid_1 = np.meshgrid(np.linspace(0.0, 1.0, self.resolution[0] * self.resolution[2] + 1), np.linspace(0.0, 1.0, self.resolution[1] + 1), indexing="ij")
        self.grid = np.stack([grid_0, grid_1], axis=0)

    def set_vertices(self, vertices):
        """Updates the polygon and related attributes."""
        self.vertices = np.copy(vertices)
        clipped_vertex_count, self.clipped_vertices = clip_polygon(self.vertices, self.vertices.shape[0] + 1)
        self.clipped_vertices = self.clipped_vertices[0:clipped_vertex_count, :]
        self.prepared_polygon = Polygon(self.vertices)

    def on_click(self, position, button):
        self.on_mouse_move(position, (0, 0), {button})

    def on_mouse_move(self, position, motion, held_buttons):
        """Updates the polygon."""
        if "lmb" not in held_buttons:
            return
        position_data = np.asarray(self.polygon_axes.transData.inverted().transform(position))
        if np.linalg.norm(position_data) >= 1.0:
            return
        vertices = np.copy(self.vertices)
        # Find the nearest vertex
        distances = np.linalg.norm(vertices[:, 0:2] - position_data[np.newaxis, :], axis=1)
        vertex_index = np.argmin(distances)
        # Move it
        vertices[vertex_index, :] = np.concatenate([position_data, [np.sqrt(1.0 - position_data @ position_data)]])
        # Update
        self.set_vertices(vertices)
        self.draw_polygon()
        return "redraw_canvas"

    def draw(self):
        draw_title(self.title)
        slide_axes = pyplot.gca()

        # Plot an arrow between the two spaces
        pyplot.arrow(7.0, 4.0, 1.0, 0.0, head_width=0.2, facecolor="k")

        # Create axes for primary sample space
        self.pss_axes = slide_axes.inset_axes((1.5, 1.5, 5.0, 5.0), transform=slide_axes.transData,
                                                  aspect="equal", frameon=True)
        pyplot.gcf().add_axes(self.pss_axes)
        self.pss_axes.set_xlim(0.0, 1.0)
        self.pss_axes.set_ylim(0.0, 1.0)
        self.pss_axes.set_xticks([])
        self.pss_axes.set_yticks([])
        # Plot the samples
        pyplot.scatter(self.random_numbers[:, 0], self.random_numbers[:, 1], color="k", s=6.0, zorder=100.0)
        # Plot the grid
        for x in range(self.resolution[0]):
            for y in range(self.resolution[1]):
                if x % 2 == y % 2:
                    continue
                inner = self.grid[0:2, x * self.resolution[2]:(x + 1) * self.resolution[2] + 1, y]
                outer = self.grid[0:2, x * self.resolution[2]:(x + 1) * self.resolution[2] + 1, y + 1]
                outline = np.concatenate([inner, outer[:, ::-1]], axis=1)
                pyplot.fill(outline[0, :], outline[1, :], color=3 * [0.75], linewidth=0.0)
        # Create axes for the polygon
        self.polygon_axes = slide_axes.inset_axes((8.5, 1.0, 6.5, 6.5), transform=slide_axes.transData,
                                                  aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.polygon_axes)
        self.draw_polygon()

    def draw_polygon(self):
        self.polygon_axes.clear()
        self.polygon_axes.set_xlim(-1.05, 1.05)
        self.polygon_axes.set_ylim(-1.05, 1.05)
        self.polygon_axes.set_xticks([])
        self.polygon_axes.set_yticks([])
        pyplot.sca(self.polygon_axes)
        # Plot the horizon and the polygon outline
        x = np.linspace(0.0, 1.0, 257)
        pyplot.plot(np.cos(2.0 * np.pi * x), np.sin(2.0 * np.pi * x), color=horizon_color)
        outline = sample_edges(self.clipped_vertices, True)
        pyplot.plot(outline[:, 0], outline[:, 1], color="k")
        # Plot samples
        samples = self.prepared_polygon.sample(self.random_numbers.T)
        pyplot.scatter(samples[0, :], samples[1, :], color="k", s=2.0, zorder=100.0)
        # Plot a grid
        grid_samples = self.prepared_polygon.sample(self.grid)
        for x in range(self.resolution[0]):
            for y in range(self.resolution[1]):
                if x % 2 == y % 2:
                    continue
                inner = grid_samples[0:2, x * self.resolution[2]:(x + 1) * self.resolution[2] + 1, y]
                outer = grid_samples[0:2, x * self.resolution[2]:(x + 1) * self.resolution[2] + 1, y + 1]
                outline = np.concatenate([inner, outer[:, ::-1]], axis=1)
                pyplot.fill(outline[0, :], outline[1, :], color=3 * [0.75], linewidth=0.0)


class LTCResultsSlide(ImageComparisonSlide):
    """Shows renderings that motivate the need for LTC sampling."""

    def __init__(self, use_ltc):
        super().__init__()
        experiment_dir = "../vulkan_renderer/data/experiments"
        suffix = "2spp" if use_ltc else "diffuse_only_1spp"
        image_name = [name for name in listdir(experiment_dir) if re.match(r"roughness_planes_lambertian_%s_.*\.png" % suffix, name)][0]
        self.images = [imread(path.join(experiment_dir, image_name))]
        self.labels = ["MIS: LTC + projected solid angle sampling" if use_ltc else "Projected solid angle sampling"]
        self.left_bottom = 1.0, 0.8
        self.image_width = 14.0
        self.offset = 5.1, 0.0
        self.half_magnifier_extent = 24, 24
        self.spp = 2 if use_ltc else 1
        self.notes = ["Diffuse+specular\nGood everywhere at 2 spp\nDifferent roughness"]

    def draw(self):
        super().draw()
        draw_title("Planes of varying roughness (%d spp)" % self.spp)


class LTCSlide(Slide):
    """Visualizes sampling of LTCs."""

    def __init__(self):
        super().__init__()
        self.step_count = 2
        self.slide_axes = self.polygon_axes = None
        self.yaw, self.pitch = -0.1 * np.pi, 0.08 * np.pi
        self.projection = get_euler_projection(self.yaw, self.pitch)
        shading_pitch = -0.6 * np.pi
        self.world_to_shading_space = np.asarray([
            [np.cos(shading_pitch), -np.sin(shading_pitch), 0.0],
            [np.sin(shading_pitch), np.cos(shading_pitch), 0.0],
            [0.0, 0.0, 1.0],
        ])
        self.shading_to_cosine_space = np.asarray([[0.782347, 0.000000, -0.134090], [0.000000, 1.000000, 0.000000], [0.422520, 0.000000, 0.262691]]).T
        self.world_to_cosine_space = self.shading_to_cosine_space @ self.world_to_shading_space
        height = 0.5
        self.polygon_world_space = np.asarray([
            [-0.5, -0.6, height],
            [0.3, -0.6, height],
            [0.3, -0.1, height],
            [-0.5, -0.1, height],
        ])
        self.polygon_world_space /= np.linalg.norm(self.polygon_world_space, axis=1)[:, np.newaxis]
        self.random_numbers = get_random_numbers()
        self.samples = Polygon(self.polygon_world_space @ self.world_to_cosine_space.T).sample(self.random_numbers.T).T
        self.step_durations = [2.0]
        # Prepare a grid of sphere samples
        self.x_min, self.x_max = -1.05, 1.05
        self.y_min, self.y_max = (-0.35, 1.05)
        dots_per_unit = 200.0
        width, height = (int((self.x_max - self.x_min) * dots_per_unit), int((self.y_max - self.y_min) * dots_per_unit))
        x, y = np.meshgrid(np.linspace(self.x_min, self.x_max, width), np.linspace(self.y_min, self.y_max, height))
        z = np.sqrt(1.0 - x * x - y * y)
        self.view_space_grid = np.stack([x, y, z], axis=2)
        self.world_space_grid = np.einsum("ij, ...j-> ...i", self.projection.matrix.T, self.view_space_grid)
        self.notes = ["Sample cos\nTransform poly+samples", "Warped samples\nLTC good fit\nSpecular IS"]

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.projection = get_euler_projection(self.yaw, self.pitch)
            self.world_space_grid = np.einsum("ij, ...j-> ...i", self.projection.matrix.T, self.view_space_grid)
            self.draw_polygon()
            return "redraw_canvas"

    def animate(self):
        if self.lerp_ongoing:
            self.draw_polygon()
            return "redraw_canvas"

    def draw(self):
        draw_title("Linearly transformed cosines [Heitz et al. 2016]")
        pyplot.text(8.25, 0.9, "cosine distribution", ha="left", va="bottom")
        if self.step >= 1:
            pyplot.text(8.25, 0.9, "linearly transformed ", ha="right", va="bottom")
        self.slide_axes = pyplot.gca()
        self.polygon_axes = self.slide_axes.inset_axes((2.0, 1.5, 12.0, 6.0), transform=self.slide_axes.transData, aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.polygon_axes)
        self.draw_polygon()

    def draw_polygon(self):
        self.polygon_axes.clear()
        self.polygon_axes.set_xlim(self.x_min, self.x_max)
        self.polygon_axes.set_ylim(self.y_min, self.y_max)
        self.polygon_axes.set_xticks([])
        self.polygon_axes.set_yticks([])
        pyplot.sca(self.polygon_axes)
        # Plot the hemisphere
        lerp_factor = np.linspace(0.0, 1.0, 257)
        horizon = np.vstack([np.cos(lerp_factor * 2.0 * np.pi), np.sin(lerp_factor * 2.0 * np.pi), np.zeros_like(lerp_factor)]).T
        self.projection.plot_curve(horizon, True, False, color=horizon_color)
        pyplot.plot(np.cos(lerp_factor * np.pi), np.sin(lerp_factor * np.pi), color=horizon_color)
        # Define the current transform and apply it
        used_world_to_cosine_space = smooth_lerp(self.world_to_cosine_space, self.world_to_shading_space, 1.0 - (self.step_lerp if self.step == 0 else 1.0))
        polygon_world_space = self.polygon_world_space @ self.world_to_cosine_space.T
        polygon_world_space = polygon_world_space @ np.linalg.inv(used_world_to_cosine_space).T
        polygon_world_space /= np.linalg.norm(polygon_world_space, axis=-1)[..., np.newaxis]
        # Plot the transformed horizon
        horizon = horizon @ np.linalg.inv(used_world_to_cosine_space).T
        horizon /= np.linalg.norm(horizon, axis=1)[:, np.newaxis]
        self.projection.plot_curve(horizon, True, False, color=horizon_color)
        # Plot the polygon outline
        outline = sample_edges(polygon_world_space, True)
        self.projection.plot_curve(outline, True, False, color=solid_angle_line_color)
        # self.projection.plot_curve(outline, False, True, color=solid_angle_line_color, linestyle="--")
        # Plot the samples
        samples = np.einsum("ij, ...j -> ...i", np.linalg.inv(used_world_to_cosine_space), self.samples)
        samples /= np.linalg.norm(samples, axis=1)[:, np.newaxis]
        samples = self.projection.transform(samples)
        mask = samples[:, 2] > 0.0
        pyplot.scatter(samples[mask, 0], samples[mask, 1], color="w", edgecolors="k", s=12.0, zorder=100.0)
        # Shade the solid angle using rasterization
        cosine_space = np.einsum("ij, ...j-> ...i", used_world_to_cosine_space, self.world_space_grid)
        cosine_space_length = np.linalg.norm(cosine_space, axis=-1)
        ltc_density = np.maximum(0.0, cosine_space[..., 2]) * np.linalg.det(used_world_to_cosine_space) / cosine_space_length ** 4
        ltc_density_rgb = cm.get_cmap("magma")(ltc_density / np.nanmax(ltc_density))[..., 0:3]
        ltc_density_rgb = np.where(np.isnan(ltc_density)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
        ltc_density_rgb = np.where((cosine_space[..., 2] < 0.0)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
        # Desaturate parts outside the solid angle of the polygon
        mask = np.ones_like(self.world_space_grid[..., 0], dtype=bool)
        for edge in range(polygon_world_space.shape[0]):
            normal = np.cross(polygon_world_space[edge], polygon_world_space[(edge + 1) % polygon_world_space.shape[0]])
            mask &= (np.einsum("...i, i -> ...", self.world_space_grid, normal) >= 0.0)
        desaturation = 0.6
        ltc_density_rgb = np.where(mask[:, :, np.newaxis].repeat(3, 2), ltc_density_rgb, (1.0 - desaturation) * ltc_density_rgb + desaturation)
        pyplot.imshow(ltc_density_rgb, extent=(self.x_min, self.x_max, self.y_min, self.y_max), zorder=-1000.0, origin="lower", interpolation="bilinear")


def maximal_error_slide(step):
    """Last slide of the presentation."""
    draw_title("Error analysis")
    draw_bullets(["",
                  "Fix iteration count to 2",
                  r"Worst case error in $\xi_0$:  $1.8\cdot10^{-5}$",
                  r"99th percentile of error in $\xi_0$:  $4.6\cdot10^{-15}$",
                  r"$\rightarrow$ unbiased"][0:(4 if step <= 0 else 5)])


def conclusions_slide(step):
    draw_title("Conclusions")
    draw_bullets([
        "Iterative methods on GPU: Better than their reputation",
        "Stability matters",
        "Put it into your renderer now",
        "Code at https://momentsingraphics.de/Siggraph2021.html",
        "Thanks!"
    ][0:[1, 2, 4, 6][step]], 1.0, 6.0)


def short_conclusions_slide(step):
    draw_title("Conclusions")
    draw_bullets([
        "Put it into your renderer now",
        "Code at https://momentsingraphics.de/Siggraph2021.html",
        "Thanks!"
    ][0:[2, 3][step]], 1.0, 5.0)


if __name__ == "__main__":
    pyplot.style.use("../Data/slides.mplstyle")
    talk = "long"
    presentation = PresentationWithFooter("../Graphics/polygon/kit_logo/kit_logo.png", 128.0 * 16.0 / 1920.0, [0] if talk != "fast" else [])
    if talk == "long":
        presentation.slides += [
            FunctionSlide(title_slide),
            RendererSlide(-1, False),
            MotivationSlide("area"),
            MotivationResultSlide(1),
            MotivationSlide("solid"),
            MotivationResultSlide(2),
            MotivationSlide("projected"),
            MotivationResultSlide(3),
            RendererSlide(49, True),
            GeometricConstructionsSlide(),
            AreaComputationSlide(),
            GeometricConstructionsSlide(5, False, 2, False),
            AzimuthSamplingSlide(),
            InitializationSlide(),
            IterationSlide(),
            FunctionSlide(maximal_error_slide, 2),
            SampleMappingSlide(),
            LTCResultsSlide(False),
            MotivationSlide("ltc"),
            LTCSlide(),
            LTCResultsSlide(True),
            RendererSlide(16, True, True, "Numerical stability"),
            FunctionSlide(conclusions_slide, 4),
        ]
    elif talk == "short":
        presentation.slides += [
            FunctionSlide(title_slide),
            RendererSlide(-1, False),
            MotivationSlide("projected"),
            MotivationResultSlide(3),
            GeometricConstructionsSlide(6, True, -1, True),
            GeometricConstructionsSlide(5, True, -1, True),
            AzimuthSamplingSlide(True),
            # InitializationSlide(),
            IterationSlide(True),
            # MotivationSlide("ltc"),
            LTCSlide(),
            LTCResultsSlide(True),
            FunctionSlide(short_conclusions_slide, 2),
        ]
    elif talk == "fast":
        presentation.slides += [
            FastMotivationSlide(),
            LTCSlide(),
            RendererSlide(-1, False),
        ]
    # export_presentation_pdf(presentation, "../Data/siggraph_slides.pdf", False)
    slideshow = Slideshow(presentation, 0, 0)
    slideshow.start()
